part of chronosgl;

class Instancer {
  
  Mesh mesh;
  int count;
  bool enabled=true;
  RenderingContext gl;
  
  ShaderObject instancedShader;
  
  Float32List transforms;
  
  Instancer(this.mesh, int count) {
    gl = ChronosGL.globalGL;
    setCount(count);
    instancedShader = createInstancedShader();
  }
  
  void setCount( int count) {
    this.count = count;
    transforms = new Float32List(count*16);
  }
  
  void setTransformFromSpatial(int position, Spatial spatial) {
    transforms.setAll(position*16, spatial.transform.array);
  }
  
  ShaderObject createInstancedShader() {
    ShaderObject shaderObject = new ShaderObject("Instanced");
    
    shaderObject.vertexShader = """
        precision mediump float;
        attribute vec3 aVertexPosition;
        
        attribute mat4 uTMatrix;
        uniform mat4 uMVMatrix;
        uniform mat4 uPMatrix;

        varying vec3 vColor;
        
        void main(void) {
          vColor = vec3( sin(aVertexPosition.x)/2.0+0.5, cos(aVertexPosition.y)/2.0+0.5, sin(aVertexPosition.z)/2.0+0.5);
          gl_Position = uPMatrix * uMVMatrix * uTMatrix * vec4(aVertexPosition, 1.0);
        }
        """;
    
    shaderObject.fragmentShader = """
        precision mediump float;
        
        varying vec3 vColor;

        void main(void) {
          gl_FragColor = vec4( vColor, 1. );
        }
        """;
    
    shaderObject.vertexPositionAttribute = "aVertexPosition"; 
    shaderObject.perpectiveMatrixUniform = "uPMatrix";
    shaderObject.modelViewMatrixUniform = "uMVMatrix";
    
    return shaderObject;
  }
  
  void draw(ShaderProgram program, Matrix4 mvMatrix) {
    
    AngleInstancedArrays extension = ChronosGL.globalGL.getExtension("ANGLE_instanced_arrays"); // Vendor prefixes may apply!
    
    mesh.bindBuffers(program);
    mesh.bindTextures(program);
    
    /*
    if( program.shaderObject.colorUniform != null) {
      program.colorUniform.setValue3fv(mesh.color);
    }
    */
    
    Buffer transformsBuffer = gl.createBuffer();

    // Bind the instance position data
    gl.bindBuffer(ARRAY_BUFFER, transformsBuffer);
    int uTMatrixLocation = program.getAttributeLocation("uTMatrix");
    gl.enableVertexAttribArray(uTMatrixLocation);
    gl.vertexAttribPointer(uTMatrixLocation, 16, FLOAT, false, 0, 0);
    extension.vertexAttribDivisorAngle(uTMatrixLocation, 1); // This makes it instanced!
    
    // Bind the instance color data
    //gl.bindBuffer(ARRAY_BUFFER, colorBuffer);
    //gl.enableVertexAttribArray(colorLocation);
    //gl.vertexAttribPointer(colorLocation, 4, gl.FLOAT, false, 16, 0);
    //extension.vertexAttribDivisorAngle(colorLocation, 1); // This makes it instanced!
    

    gl.uniformMatrix4fv(program.modelViewMatrixUniform, false, mvMatrix.array);
    
    if( mesh.vertexIndexBuffer == null) {
      extension.drawArraysInstancedAngle(TRIANGLES, 0, mesh.numItems, this.count);
    } else  {
      gl.bindBuffer(ELEMENT_ARRAY_BUFFER, mesh.vertexIndexBuffer);
      extension.drawElementsInstancedAngle(TRIANGLES, mesh.numItems, ChronosGL.useElementIndexUint ? UNSIGNED_INT : UNSIGNED_SHORT, 0, this.count);
    }
    
    gl.disableVertexAttribArray(uTMatrixLocation);

  }
  
}

